(* --------------------------------------------------------------------
 * (c) Copyright 2014--2015 INRIA.
 *
 * You may distribute this file under the terms of the CeCILL-B license
 * -------------------------------------------------------------------- *)

Require Import ssreflect ssrfun ssrbool eqtype ssrnat seq choice fintype.
Require Import finfun div binomial bigop prime ssralg poly ssrnum ssrint.
Require Import intdiv ssrcomplex Rstruct Cstruct Canalysis mpolytrans Exp_prop.
Require Import mxpoly rat Rtrigo_def.


Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

Open Scope ring_scope.
Import GRing.Theory.
Import Num.Theory.

Section Preliminaries.

Hypothesis algebraic_e : algebraicOver  ratr (exp 1)%:C.

(* construct the polynomial by removing zero roots*)
Lemma ex_A_e :
  exists P : {poly int}, (P != 0) && root (map_poly intr P) (exp 1)%:C &&
    (P`_0 != 0).
Proof.
have e_n0 : (exp 1)%:C != 0.
  by rewrite eq_complex /= eq_refl andbT gtr_eqF //; apply/RltP/exp_pos.
by case: (poly_caract_int algebraic_e e_n0) => P /andP [pP _]; exists P.
Qed.

(* Variables definitions for the complex polynomials *)

Definition A_e := xchoose (ex_A_e).
Definition n := (size A_e).-1.
Definition k := A_e`_0.
Definition gamma i := A_e`_(i.+1).
Definition alpha i := (i.+1%:R : complexR).

Lemma n_neq0 : n != 0%N.
Proof.
rewrite /n.
move: (xchooseP ex_A_e).
rewrite -/A_e => /andP [/andP [Hdiff0 Hroot] Hneq0].
rewrite -lt0n -(ltn_add2l 1) add1n -subn1.
rewrite subnKC; last by rewrite size_poly_gt0.
have H :  (lead_coef A_e)%:~R != 0 :> complexR.
  by rewrite intr_eq0 lead_coef_eq0.
rewrite -(@size_map_poly_id0 _ _ (intr : int -> complexR)) //.
apply: (@root_size_gt1 _ (exp 1)%:C)=>//.
by rewrite -lead_coef_eq0 lead_coefE coef_map_id0 // size_map_poly_id0.
Qed.

Lemma k_neq0 : k != 0.
Proof.
rewrite /k.
move: (xchooseP ex_A_e).
by rewrite -/A_e => /andP [/andP [Hdiff0 Hroot] Hneq0].
Qed.

Lemma alpha_neq0 :
  forall i, alpha i != 0.
Proof. by move=> i; rewrite /alpha pnatr_eq0 -lt0n ltn0Sn. Qed.

Lemma eq_caract_e :
    k%:~R + \sum_(i < n) (gamma i)%:~R * (exp ((i.+1)%:R))%:C = 0.
Proof.
have : \sum_(i < n) (gamma i)%:~R * (exp i.+1%:R)%:C =
            \sum_(i < n) (gamma i)%:~R * ((exp 1)%:C ^+ (i.+1)) => [|->].
  have H : forall i , ((exp ((i.+1)%:R))%:C) =  ((exp 1)%:C ^+ (i.+1)).
    by move=> i; rewrite -expRX rmorphX.
  by apply: congr_big => //; move=> i T; rewrite H.
set e := (exp 1)%:C; move: (xchooseP ex_A_e).
rewrite /k /gamma -/A_e => /andP [/andP [Hdiff0 Hroot] Hneq0].
have -> : (A_e`_0)%:~R = (A_e`_0)%:~R * e ^+ 0  by rewrite expr0 mulr1.
rewrite -(big_ord_recl n (fun i => (A_e`_i)%:~R * e ^+ i)) .
rewrite -(horner_poly (n.+1) (fun i => (A_e`_i)%:~R) e).
apply/rootP.
suff ->: \poly_(i < n.+1) ((A_e`_i)%:~R : complexR) = map_poly intr A_e by [].
by rewrite -polyP => i; rewrite 2!coef_poly /n prednK // size_poly_gt0.
Qed.

Lemma alpha_Cint : forall i0 : nat, alpha i0 \in Cint.
Proof. by move=> i; rewrite /alpha; rewrite -natzc; apply: Cint_int. Qed.
Hint Resolve alpha_Cint.

Lemma e_transcendental_aux : False.
Proof.
have o0 : (1%N != 0%N) by [].
have mc := (main_contradiction n_neq0 k_neq0 o0 alpha_neq0).
have T_Cint: T n 1 alpha \is a polyOver Cint.
  by apply: rpred_prod => i _; rewrite polyOverXsubC alpha_Cint.
apply: (mc gamma T_Cint).
  have t := eq_caract_e.
  have q : forall i, (0 <= i < n)%N -> 
         (gamma i)%:~R * Cexp (alpha i) = (gamma i)%:~R * (exp i.+1%:R)%:C.
    move => i _; congr (_ * _).
    rewrite Cexp_exp; last by rewrite Creal_Cint ?alpha_Cint.
    rewrite /alpha; congr ((exp _)%:C).
    by elim:i =>// [i I]; rewrite -addn1 !natrD -I; case (i.+1%:R : complexR).
  by rewrite (eq_big_nat 0 +%R q) -[RHS]t big_mkord.
apply: rpred_sum => i _; rewrite -mulr_natr exp1n mulr1.
by rewrite rpredM ?rpred_horner ?G_Cint ?Cint_int.
Qed.

End Preliminaries.

Section Trans_e.

Theorem e_transcendental :  ~ (algebraicOver ratr (exp 1)%:C).
Proof. exact e_transcendental_aux. Qed.

End Trans_e.
