(* --------------------------------------------------------------------
 * (c) Copyright 2014--2015 INRIA.
 *
 * You may distribute this file under the terms of the CeCILL-B license
 * -------------------------------------------------------------------- *)

Require Import ssreflect ssrfun ssrbool eqtype ssrnat seq choice fintype.
Require Import finfun div binomial bigop prime ssralg poly ssrnum ssrint.
Require Import finset intdiv perm fingroup.
Require Import ssrcomplex Rstruct Cstruct Exp_prop Rtrigo_def.
Require Import Rtrigo1 mxpoly rat tuple freeg. 
Require Import mpoly bigenough.

Require Import mpolytrans.

Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

Open Scope ring_scope.
Import GRing.Theory BigEnough.
Import Num.Theory.

Section Preliminary.

(* move to poly ? *)
Lemma ccoeffn0_Pn0 (F : fieldType)  (P :{poly F}): (P`_0 != 0) -> (P != 0).
Proof. by apply/contra =>/eqP ->; rewrite coef0. Qed.

(* eq caract i * pi *)

Hypothesis algebraic_pi : algebraicOver ratr (PI)%:C.

Lemma algebraic_ipi :
    algebraicOver ratr ('i * (PI)%:C).
Proof.
apply: algebraic_mul=>//.
exists ('X^2 + 1%:P).
  by rewrite -size_poly_eq0 size_addl size_polyXn ?size_poly1.
apply/rootP; rewrite rmorphD  hornerD rmorphX /= map_polyX map_polyC hornerC.
by rewrite hornerXn sqr_i /= rmorph1 addNr.
Qed.

Lemma ipi_n0 : 'i * PI%:C != 0.
Proof.
rewrite eq_complex /= !mulr0 mul0r subr0 eqxx mul1r add0r gtr_eqF //.
by apply/RltP; apply PI_RGT_0.
Qed.

Definition BpiI := projT1 (sigW (poly_caract_int algebraic_ipi ipi_n0)).
Notation Bpi := (map_poly (intr : int -> complexR) BpiI).

Lemma rootBpi: root Bpi ('i * (PI)%:C).
Proof. by rewrite /BpiI; case: sigW => P /= /andP [/andP [/andP []]]. Qed.

Lemma Bpi_0 : Bpi`_0 != 0.
Proof.
rewrite /BpiI; case:(sigW _); simpl projT1=> P /andP [/andP []] _ cn0 _.
by rewrite coef_map intr_eq0.
Qed.

Definition betaseq := projT1 (closed_field_poly_normal Bpi).

Lemma BpiE :
   Bpi = (lead_coef Bpi) *: \prod_(z <- betaseq) ('X - z%:P).
Proof.
rewrite /betaseq; case: closed_field_poly_normal => b /= ->; congr (_ *: _).
rewrite lead_coefZ; rewrite [X in _ * X](_ : _ = 1) ?mulr1//.
by apply/monicP/rpred_prod => a _; apply/monicP; rewrite lead_coefXsubC.
Qed.

Lemma leadBpi : lead_coef Bpi != 0.
Proof. by rewrite lead_coef_eq0 ccoeffn0_Pn0 ?Bpi_0. Qed.

Lemma iPi_in_betaseq : 'i * PI%:C \in betaseq.
Proof. 
have := rootBpi; rewrite BpiE /root. 
rewrite hornerZ mulf_eq0 -[_.[_] ==0]/(root (\prod_(z <- _)('X - z%:P)) _).
by rewrite root_prod_XsubC (negbTE leadBpi).
Qed.

Notation s := (size betaseq).-1.

(* tuple_beta *)
Lemma size_beta : size betaseq = s.+1.
Proof. 
rewrite prednK // -has_predT; apply/hasP; exists ('i * PI%:C)%R => //.
exact: iPi_in_betaseq.
Qed.

Lemma size_beta_eqb : size betaseq == s.+1.
Proof. by rewrite size_beta. Qed.

Lemma size_Bpi : size Bpi = s.+2.
Proof. 
by rewrite BpiE size_scale ?leadBpi // size_prod_XsubC size_beta. Qed.

Definition betatuple := Tuple size_beta_eqb.

Fact size_Cexpbeta : size [seq Cexp i | i <- betaseq] == s.+1.
Proof. by rewrite  size_map size_beta. Qed.

Definition Cexpbetatuple := Tuple size_Cexpbeta.

Lemma prod_betaseq_Cexp :
    \prod_(b <- betaseq) (1 + (Cexp b)) = 0.
Proof.
move: iPi_in_betaseq; rewrite mulrC =>/perm_to_rem Heq.
by rewrite (eq_big_perm _ Heq) /= big_cons Euler mul0r.
Qed.


Lemma prod_betaseq_poly :
    \prod_(i <- betaseq) (1 + (Cexp i)) =
    (\prod_(i <- betaseq) ('X +  (Cexp i)%:P)).[1].
Proof.
rewrite horner_prod; apply: eq_big_seq => x Hx.
by rewrite hornerD hornerX hornerC.
Qed.

Local Notation "''s_' ( n , k )" := (@mesym n _ k).


Lemma viete_poly n (cs : n.+1.-tuple complexR) :
  (\prod_(c <- cs) ('X - c%:P)) =
  \poly_(i < n.+2) ('s_(n.+1,(n.+1 -i)).@[cs] * ((-1)^+(n.+1-i))).
Proof.
rewrite -polyP => i.
have Hsize : size (\prod_(c <- cs) ('X - c%:P)) = n.+2.
  by rewrite size_prod_XsubC size_tuple.
(case: (boolP (i <= n.+1)%N); rewrite -ltnS)=> [Hin | /negbTE Hin].
  rewrite coef_poly Hin.
  pose j := rev_ord (Ordinal Hin).
  have -> : i = (n.+1 - j)%N by rewrite /= subSS subKn.
  by rewrite mroots_coeff subKn; [rewrite mulrC|rewrite -ltnS].
rewrite coef_poly  Hin; apply/(leq_sizeP _ n.+2); rewrite ?Hsize //.
by move/negbT: Hin; rewrite -leqNgt.
Qed.

Lemma viete_poly_pos n (cs : n.+1.-tuple complexR) :
    (\prod_(c <- cs) ('X + c%:P)) =
    \poly_(k < n.+2) 's_(n.+1,(n.+1 -k)).@[cs].
Proof.
pose csm := [tuple of (map (fun x => -x) cs)].
have -> : \prod_(c <- cs) ('X + c%:P) = \prod_(c <- csm) ('X - c%:P).
  by rewrite big_map; apply: eq_bigr => i _; rewrite -polyC_opp opprK.
rewrite viete_poly -polyP => i.
rewrite !coef_poly; case: ifP => // H.
move: (n.+1 - i)%N  => {i H} i.
rewrite  big_distrl  big_seq_cond [X in _ = X]big_seq_cond.
apply: eq_bigr => m; rewrite andbT => Hm.
rewrite mcoeff_mesym.
case H : (mechar i m); last by rewrite /= !mul0r.
rewrite !mul1r /mmap1; move: H.
rewrite /mechar => /andP[/eqP Hmdeg /forallP H].
rewrite (eq_bigr (fun x => (tnth cs x ^+ m x) * (-1) ^+ m x)).
rewrite big_split /= prodrXr.
  move: Hmdeg; rewrite /mdeg big_tuple (eq_bigr (fun x => m x)) => Hmdeg //.
  by rewrite Hmdeg mulrC [X in X = _]mulrC signrMK.
by move=> j _; rewrite tnth_map -[- tnth cs j]mulrN1 exprMn.
Qed.

Lemma horner_poly1  n (E : nat -> complexR):
    (\poly_(i < n) E i).[1] = \sum_(i < n) E i.
Proof.
by rewrite horner_poly; apply: eq_big => //  i _; rewrite expr1n mulr1.
Qed.

Fact Cexpbetatuple_sym (sf : {set 'I_s.+1}) :
    \prod_(i in sf) Cexpbetatuple`_i = Cexp(\sum_(i in sf) betatuple`_i).
Proof.
have m : {morph Cexp : x y / x + y >-> x * y} by move=> x y /=; rewrite CexpRD.
rewrite (big_morph Cexp m Cexp0).
by apply: eq_bigr => [[]] i H _; rewrite (nth_map 0) //= size_beta.
Qed.

Fact polysym_Cexp (k : nat) :
    's_(s.+1,k).@[Cexpbetatuple] =
    \sum_(h : {set 'I_s.+1} | #|h| == k) (\prod_(i in h) Cexpbetatuple`_i).
Proof.
rewrite (big_morph _ (mevalD Cexpbetatuple) (meval0 _)).
apply: eq_bigr => h h_mono.
suff meval11 : 1.@[Cexpbetatuple] = 1.
  rewrite (big_morph _ (mevalM Cexpbetatuple) meval11).
  by apply eq_bigr => i _; rewrite mevalX -tnth_nth.
rewrite /meval /mmap msupp1 big_seq1 mmap11 mulr1 -mpolyC1 mcoeffC.
by rewrite eq_refl  mulr1.
Qed.

Fact prod_betaseq_poly2 :
  \prod_(i <- betaseq) (1 + (Cexp i)) =
  (\prod_(i <- Cexpbetatuple) ('X + i%:P)).[1].
Proof.
rewrite prod_betaseq_poly.
suff :  (\prod_(i <- betatuple) ('X + ( Cexp i)%:P)) =
           (\prod_(i <- Cexpbetatuple) ('X + i%:P)) => [-> //|].
do 2! rewrite big_tuple; apply: eq_big => // i _.
 by rewrite -(tnth_map Cexp) !(tnth_nth 0).
Qed.

Fact viete_Cexpbetatuple :
    (\prod_(i <- Cexpbetatuple) ('X + i%:P)).[1] =
    ((\sum_(k < s.+2) ( ('s_(s.+1,(s.+1 -k)).@[Cexpbetatuple]))) : complexR).
Proof. by rewrite viete_poly_pos horner_poly1. Qed.

Lemma polysym_Cexpa :
    \prod_(b <- betaseq) (1 + (Cexp b)) =
    \sum_(k < s.+1)
      (\sum_(h : {set 'I_(s.+1)} | #|h| == k.+1)
        Cexp(\sum_(i0 in h) betaseq`_i0) ) + 1.
Proof.
rewrite prod_betaseq_poly2 viete_Cexpbetatuple big_ord_recr subnn.
rewrite [X in X = _]/= mesym0E {2}/meval /mmap  msupp1.
rewrite big_seq1 mmap11 mulr1 -mpolyC1 mcoeffC eq_refl mulr1.
congr (_ + 1); rewrite (reindex_inj (@rev_ord_inj s.+1)).
apply: eq_bigr => i _.
rewrite [X in X = _]/= (subKn (ltn_ord _)) polysym_Cexp.
by apply: eq_bigr => h _; apply: Cexpbetatuple_sym.
Qed.

(* obtention des alpha *)
(* par rang *)


Definition alpha'_setk (k : 'I_s.+1) :=
    [set h in {set 'I_s.+1} | #|h| == k.+1 ].

Definition pre_alpha' (k : 'I_s.+1) :=
    [seq (\sum_(i in (pred_of_set h)) betaseq`_(nat_of_ord i)) |
      h in (alpha'_setk k)].

Fact alpha_beta_eqk (k : 'I_s.+1) :
    (\prod_(h : {set 'I_(s.+1)} | #|h| == k.+1)
      ('X - (\sum_(i0 in h) betaseq`_i0)%:P) ) =
    \prod_(i <- (pre_alpha' k)) ('X - i%:P).
Proof. 
 by rewrite big_map big_filter; apply: eq_bigl => h; rewrite /= inE.
Qed.

Definition alpha'_tuplek (k : 'I_s.+1) :=
    in_tuple [seq (\sum_(i in (pred_of_set h))
      ('X_i : {mpoly complexR[s.+1]})) | h in (alpha'_setk k)].

Definition alpha'beta_tuplek (k : 'I_s.+1) :=
    (map (meval betatuple) (alpha'_tuplek k)).

Fact alpha_beta_tuplek (k : 'I_s.+1) : pre_alpha' k = alpha'beta_tuplek k.
Proof.
rewrite /pre_alpha' /alpha'beta_tuplek.
set seqa := [seq _ | _ in _].
set seqb := [seq j.@[betatuple] | j <- alpha'_tuplek k].
have Hs : size seqa = size seqb.
  by rewrite /seqb /seqa ![X in _ = X]size_map size_map.
apply: (@eq_from_nth _ 0) => // j Hj.
rewrite (nth_map 0);last first.
  by  move:Hj; rewrite Hs size_map.
rewrite /seqa !(nth_map set0); move: Hj; rewrite /seqa size_map //.
  rewrite raddf_sum => Hj; apply: eq_bigr => i _ /=.
  by rewrite -/meval mevalX (tnth_nth 0).
Qed.

Fact alpha_beta_mpoly (k : 'I_s.+1) :
    (\prod_(h : {set 'I_(s.+1)} | #|h| == k.+1)
      ('X - (\sum_(i0 in h) betatuple`_i0)%:P) ) =
    (\prod_(h : {set 'I_(s.+1)} | #|h| == k.+1)
      ('X - ((\sum_(i0 in h) 'X_i0).@[betatuple]%:P))).
Proof.
apply: eq_bigr => i _; congr (fun x => 'X - x%:P).
by rewrite raddf_sum; apply: eq_bigr => j _; rewrite /= mevalX (tnth_nth 0).
Qed.

Lemma mcoeff_mpolyE nu (m : 'X_{1..nu}) (R1 : ringType) (R2 : ringType)
    (p : {mpoly R1[nu]})
    (f : 'X_{1..nu} -> R2) :
    m \in msupp p ->
    (\sum_(mu <- msupp p) (f mu) *: 'X_[mu])@_m = f m.
Proof.
move=> msupp_m.
rewrite raddf_sum  (bigD1_seq m msupp_m (msupp_uniq _)).
rewrite big1 /=; last first.
  move=> i /negbTE i_neqm.
  by rewrite mcoeffZ mcoeffX i_neqm mulr0.
by rewrite addr0 mcoeffZ mcoeffX eq_refl mulr1.
Qed.

Lemma mpolyC_intr_is_additive nu : additive ((@mpolyC nu
    complexR_ringType) \o (intr)).
Proof. by apply: GRing.comp_is_additive. Qed.

Canonical mpolyC_intr_additive nu :=
Additive (mpolyC_intr_is_additive nu).

Implicit Arguments mpolyOver [n R].

Lemma mpolysym nu (p : {mpoly complexR[nu]}) :
    p \is a mpolyOver Cint -> p \is symmetric -> p != 0 ->
    {q | [&& (p == q \mPo [tuple 's_(nu, i.+1) | i < nu]),
          ((mweight q) <= msize p)%N
          & (q \is a mpolyOver Cint)]}.
Proof.
move=> /mpolyOverP Hp /msym_coeff psym p_neq0.
pose p_int := \sum_(m <- msupp p) (floorC p@_m) *: 'X_[m].
have mcoeff_p_int m : p_int@_m = (floorC p@_m).
  case: (boolP (m \in msupp p)) => Hm .
    by rewrite /p_int mcoeff_mpolyE .
  move:(Hm); rewrite -mcoeff_eq0 => /eqP ->; rewrite floorC0. 
  rewrite /p_int raddf_sum big1 // => i _;rewrite /= mcoeffZ mcoeffX.
  case: (boolP (i == m)); rewrite ?mulr0 // => /eqP Him.
  by move :Hm; rewrite Him  mulr1 -mcoeff_eq0 =>/eqP -> ; rewrite floorC0.
have p_int_sym : p_int \is symmetric.
  apply/issymP=> x; rewrite -mpolyP => m.
  by rewrite mcoeff_sym !mcoeff_p_int  psym.
have msupp_p_perm_eq : perm_eq (msupp p_int) (msupp p).
  rewrite (perm_eq_trans (msupp_sumX (msupp_uniq _ )_)) // =>  m Hm.
  move: (Hp m); rewrite /Cint unfold_in => /eqP H.
  by rewrite -(@eqr_int complexR_numDomainType) H mcoeff_eq0 negbK.
move:(sym_fundamental p_int_sym)=> [q_int [q_int_eq qweight]].
have Hperm_eq rint :
      perm_eq (msupp (mmap (fun x => (mpolyC nu(intr x : complexR)))
        (fun i => 'X_i) rint)) (msupp rint).
  rewrite /mmap (eq_bigr (fun m => ((rint@_m)%:~R) *: 'X_[m])).
    rewrite (perm_eq_trans (msupp_sumX (msupp_uniq _ )_)) // =>  m.
    by rewrite mcoeff_msupp intr_eq0.
  by move=>i _; rewrite mul_mpolyC mmap1_id.
have p_int_size : msize (p_int) = msize p.
  by rewrite  msizeE (eq_big_perm _ msupp_p_perm_eq).  
exists (mmap (fun x => (mpolyC nu (intr x : complexR)))(fun i => 'X_i) q_int).
set q := mmap _ _ _.
apply/and3P; split; last first.
    apply/mpolyOverP => m.
    rewrite /q /mmap (eq_bigr (fun m0 => ((q_int@_m0)%:~R) *: 'X_[m0])). 
      case: (boolP (m \in msupp q_int)) => Hm.
        by rewrite (mcoeff_mpolyE _ Hm); apply: Cint_int.
      rewrite raddf_sum big1 // =>  i _; rewrite /=  mcoeffZ mcoeffX.
      case: (boolP ( i == m)); rewrite ?mulr0 // mulr1=> /eqP ->.
      by move:Hm; rewrite -mcoeff_eq0=> /eqP ->.
    by move=> i _; rewrite mul_mpolyC mmap1_id.
  have msupp_q_perm_eq : perm_eq (msupp q) (msupp q_int).
    by rewrite /q; apply: Hperm_eq.      
  have -> : mweight q = mweight q_int. 
    by rewrite !mmeasureE (eq_big_perm _ msupp_q_perm_eq).
  by rewrite -p_int_size.
have -> : p = mmap (fun x : int_Ring => (x%:~R)%:MP_[nu])
         (fun i : 'I_nu => 'X_i) p_int.
  rewrite /mmap (eq_big_perm _ msupp_p_perm_eq) /= {1}[p]mpolyE.
  apply: eq_bigr => m _ .
  by rewrite mcoeff_p_int mmap1_id mul_mpolyC floorCK.
rewrite -q_int_eq.
set f := (fun x : int_Ring => _).
set mid := (fun i : _ => _).
move => {Hp psym p_neq0 mcoeff_p_int msupp_p_perm_eq p_int_size}.
move => {p_int_sym q_int_eq qweight p_int p}.
pose_big_enough i.
  rewrite /comp_mpoly (mmapE i) //.
  rewrite (raddf_sum (mmap_additive mid (mpolyC_intr_additive nu))) (mmapE i) //=.
  apply/eqP; apply: eq_bigr => m _.
  rewrite !mul_mpolyC mmapZ /= mul_mpolyC.
  case : (boolP (val m \in msupp q_int)) => Hm; last first.
    move:(Hm); rewrite -mcoeff_eq0; move /eqP -> => /=.
    rewrite /q {2}/mmap [X in (_ = X *:_)]raddf_sum /= big1.
      by rewrite !scale0r.
    move => j _; rewrite mmap1_id mul_mpolyC mcoeffZ mcoeffX.
    case:(boolP (j == m)); rewrite ?mulr0 // mulr1 => /eqP ->.
    by move:Hm; rewrite  -mcoeff_eq0  => /eqP  ->.
  rewrite /q {2}/mmap (eq_bigr (fun m0 => (q_int@_m0)%:~R *: 'X_[m0]));
        last by move=> mu _; rewrite mul_mpolyC mmap1_id.
  rewrite (mcoeff_mpolyE _ Hm) /=; congr  (_  *: _).
  rewrite  rmorph_prod; apply: eq_bigr => j _.
  rewrite rmorphX ; congr (_ ^+ _).
  rewrite !tnth_mktuple /mesym raddf_sum /=.
  apply: eq_bigr => i0 _; rewrite rmorph_prod /=; apply: eq_bigr => i1 _.
  by rewrite mmapX /mid mmap1_id.
by close.
Qed.

(* this theorem is true for any tuple, not just betatuple. *)
Lemma mesym_beta_Cint k d: 
  ((\prod_(i <- betatuple) ('X - i%:P)) *+ d \is a polyOver Cint)->
    (k <= s.+1)%N -> 's_(s.+1,k).@[betatuple] *+ d \in Cint.
Proof.
move/polyOverP=> H; rewrite -ltnS => Hk.
move: (H (s.+1 - (Ordinal Hk))%N).
by rewrite coefMn mroots_coeff -mulrnAr rpredMsign.
Qed.

Lemma  s_alpha j k :  let sa:= size(alpha'_tuplek k) in
                       ('s_(sa, j))  \mPo alpha'_tuplek k \is symmetric.
Proof.
apply: msym_comp_poly; first by apply: mesym_sym.
move=> sp; rewrite /alpha'_tuplek /=.
pose ms spe := (fun (z : set_of_finType (ordinal_finType s.+1)) =>
        [set spe zi | zi in z]
          : set_of_finType (ordinal_finType s.+1)).
have Hcancel (z : 'S_s.+1) : cancel (ms z) (ms (z^-1)%g).
  move=>x; rewrite /ms -imset_comp.
  rewrite (@eq_imset _ _ _ (fun za => za)) ?imset_id //.
    by move=> za /=; rewrite permK.
have H1 : [set (ms sp) hi | hi in alpha'_setk k] = alpha'_setk k.
  rewrite -setP => x.
  case: (boolP (#|x| == k.+1)) => [/eqP H | /negbTE H].
    rewrite in_set H eq_refl andbT inE /=.
    apply/imsetP; exists (ms (sp^-1)%g x).
      rewrite in_set inE andTb card_imset; first by rewrite H.
      by move=> y z /=; apply: perm_inj.
    by rewrite -{1}[x](Hcancel (sp^-1)%g) invgK.
  rewrite in_set H andbF.
  apply/negP => /imsetP [y Hy yeq].
  move: Hy; rewrite in_set inE andTb => Hy.
  move: H => /negP H; apply: H; rewrite yeq card_imset //.
  by move=> z1 z2; apply: perm_inj.
  rewrite -{1}H1.
  set sa := (size (alpha'_tuplek k)).
have -> :  [seq msym sp [seq \sum_(i1 in pred_of_set h) 'X_i1
                   | h in alpha'_setk k]`_(nat_of_ord i0) | i0 <- enum 'I_sa] =
                [seq \sum_(i0 in pred_of_set h) 'X_i0
                   | h <- (map (ms sp) (enum (alpha'_setk k)))].
  move=> t.
  set seqa := [seq _ | _ <- _].
  set seqb := [seq _ | _ <- _].
  have Hseq : size seqa = size seqb.
    by rewrite /seqa /seqb size_map size_enum_ord !size_map /sa size_map.
  have Hord0 : (0 < sa)%N.
    rewrite /sa /alpha'_tuplek /= size_map -cardE /alpha'_setk /=.
    by rewrite card_draws card_ord bin_gt0.
  apply : (@eq_from_nth _ 0); first by rewrite Hseq.
  move=> i0 Hi0; rewrite (nth_map set0) /=; last first.
    by move: Hseq; rewrite /seqb size_map => <-.
  rewrite (nth_map (Ordinal Hord0)); last first.
    by move: Hi0; rewrite /seqb size_map.
  rewrite nth_enum_ord; last by move: Hi0; rewrite size_map size_enum_ord.
  rewrite (nth_map set0) /=; last first.
    by move: Hseq; rewrite /seqb size_map [X in _ =X]size_map => <-.
  rewrite (nth_map set0) /=; last first.
    by move: Hseq; rewrite /seqb size_map [X in _ =X]size_map => <-.
  set h := nth _ _ _.
  rewrite raddf_sum /= big_imset /=; last first.
    by move=> z1 z2 _ _ /=; apply: perm_inj.
  by apply: eq_bigr => i1 _; rewrite /msym mmapX mmap1U.
apply: perm_eq_map.
apply: uniq_perm_eq; first by apply: enum_uniq.
  rewrite map_inj_uniq; first by apply: enum_uniq.
  apply: bij_inj; apply: (@Bijective _ _ _ (ms (sp^-1)%g)) => x.
    by rewrite Hcancel.
  by rewrite -[X in _ =X](Hcancel (sp^-1)%g) invgK.
 move=> x; rewrite mem_enum.
by apply/imsetP/mapP => [|][y Hy yeq]; exists y=> //; move: Hy; rewrite mem_enum.
Qed.

Lemma alpha'_int (j : 'I_s.+1) :
    {c : nat |
        (\prod_(a <- pre_alpha' j)('X - a%:P) *+ c \is a polyOver Cint) &&
        (c != 0%N)}.
Proof.
pose sa := size (alpha'_tuplek j).
pose m (l : 'I_sa.+1) := msize (comp_mpoly (alpha'_tuplek j) 's_(sa, l)).
pose mmax := \max_(j < sa.+1) m j.
set d := absz (lead_coef BpiI).
have d_neq0 : (d != 0)%N.
  rewrite /d absz_eq0 /BpiI lead_coef_eq0; case:(sigW _)=> p; simpl projT1.
  by move=>/andP [/andP [/andP [A _] _] _].
exists (d ^ mmax)%N; apply/andP; split; last first.
  by rewrite expn_eq0; move/negbTE: d_neq0 => ->.
apply/polyOverP => i; rewrite coefMn.
have Hsize : size (\prod_(i0 <- pre_alpha' j)('X - i0%:P)) = sa.+1.
  by rewrite size_prod_XsubC /sa  !size_map.
case : (boolP (i < sa.+1)%N) => [Hordi |]; last first.
  rewrite -ltnNge -Hsize => /leq_sizeP  /(_ i (leqnn i)) -> .
  by apply/rpredMn/rpred0.
have Hordj : (sa - i < sa.+1)%N by apply:(leq_ltn_trans (leq_subr _ _ )).
pose l := Ordinal Hordj.
have -> : i = (sa - l)%N by rewrite subKn.
have Ha : size (pre_alpha' j) == sa by rewrite alpha_beta_tuplek size_map.
pose tuple_alpha := Tuple Ha.
have -> : pre_alpha' j = tuple_alpha by [].
rewrite mroots_coeff -/sa -mulrnAr rpredMsign.
have Hs: size(map (meval betatuple)(alpha'_tuplek j)) == size(alpha'_tuplek j).
  by rewrite size_map.
have -> : tuple_alpha = Tuple Hs.
  apply: eq_from_tnth => u; rewrite !(tnth_nth 0); congr (fun x => x`_u).
  by rewrite /= alpha_beta_tuplek. 
set p := 's_(sa, l).
have -> : p.@[Tuple Hs] = (p \mPo alpha'_tuplek j).@[betatuple].
  rewrite {1}/meval /mmap raddf_sum.
  apply: eq_bigr => m' _ /=.
  rewrite mevalM mevalC; congr (fun x => p@_m' * x).
  rewrite /mmap1 rmorph_prod /=; apply: eq_bigr => i' _ /=.
  rewrite rmorphX /=; congr (fun x => x ^+ m' i').
  by rewrite !(tnth_nth 0) /= (nth_map 0) // size_tuple.
set P := comp_mpoly _ _.
have P_sym : P \is symmetric by apply: (s_alpha l j).
have P_mpolyOver : P \is a mpolyOver Cint.
  rewrite /P /comp_mpoly /mmap; apply: rpred_sum => mu _; apply: rpredM.
    have : 's_(sa,l) \is a mpolyOver Cint.
      rewrite /mesym; apply: rpred_sum => cs _.
      apply: rpred_prod => mi _.
      by apply: mpolyOverX.
    by move=> /mpolyOverP Hsym; rewrite mpolyOverC; apply: Hsym.
  rewrite /mmap1; apply: rpred_prod => u _; apply: rpredX.
  rewrite (tnth_nth 0) /= (nth_map set0); last first.
    by move: (ltn_ord u); rewrite /sa /= [X in (_ < X)%N -> _]size_map.
  by apply: rpred_sum => i0 _; apply: mpolyOverX.
case: (boolP (P == 0)) => [/eqP P_eq0 | P_neq0].
  by rewrite P_eq0 meval0 mul0rn.
move: (mpolysym P_mpolyOver P_sym P_neq0).
move=> [] Q /and3P [/eqP Qeq Qsize /mpolyOverP Q_Cint]; rewrite Qeq.
have Hmmax1 : (msize Q <= mmax)%N.
  apply: (leq_trans (leq_msize_meight _)).
  apply: (@leq_trans (m l)); first by rewrite /m.
    by rewrite /mmax; apply: leq_bigmax.
move=> {Qeq P_neq0 P_mpolyOver P_sym P Hs tuple_alpha Qsize}.
move=> {Ha p l Hordj Hordi Hsize i}.
have -> : (Q \mPo [tuple 's_(s.+1,i.+1) | i < s.+1]).@[betatuple] =
    meval [tuple 's_(s.+1,i.+1).@[betatuple] | i < s.+1] Q.
  rewrite [X in _ = X]/meval /mmap /comp_mpoly /mmap raddf_sum /=.
  apply: eq_bigr => mu _ /=; rewrite mevalM mevalC.
  congr (_ * _); rewrite /mmap1 rmorph_prod /=.
  apply: eq_bigr => i _.
  rewrite rmorphX /=; congr (_ ^+ _).
  by rewrite !(tnth_nth 0) !(nth_map ord0) -?enumT ?size_enum_ord.
rewrite {1}/meval /mmap -sumrMnl big_seq.
apply: rpred_sum => mu /nthP Hmu; move: (Hmu 0%MM) => {Hmu} [imu Hmu mue].
rewrite -mulrnAr /mmap1; apply: rpredM; first by rewrite /= Q_Cint.
suff H : (\prod_(i < s.+1) (tnth [tuple ('s_(s.+1,i0.+1)).@[betatuple] |
      i0 < s.+1] i ^+ mu i)) * (\prod_(i < s.+1) d%:R ^+ mu i) \in Cint.
  move: H; rewrite prodrXr -natrX mulr_natr.
  set mm := (\sum_(i <- _) mu i)%N => H.
  have Hmmax2 : (mm <= mmax)%N.
    apply: (leq_trans _ Hmmax1).
    rewrite /mm mmeasureE.
    apply: (@leq_trans (mdeg mu).+1).
      rewrite mdegE -[X in (X <= _)%N]addn0 -[X in (_ <= X)%N]addn1 /=.
      by rewrite leq_add2l.
    rewrite (big_nth 0%MM) big_mkord -mue.
    set imuord := Ordinal Hmu.
    have -> : imu = nat_of_ord imuord by [].
    by apply: (leq_bigmax).
  rewrite -(subnKC Hmmax2) expnD mulrnA -mulr_natr.
  apply: rpredM => //.
  by apply: Cint_Cnat; apply: Cnat_nat.
rewrite -big_split /=; apply: rpred_prod => i _.
rewrite -exprMn; apply: rpredX.
rewrite tnth_mktuple mulr_natr.
apply: mesym_beta_Cint; last by case: i.
rewrite -scaler_nat /d. 
have -> : `|lead_coef BpiI|%:R = lead_coef Bpi.
  rewrite /BpiI; case: (sigW _); simpl projT1 => p /andP [A B].
  rewrite lead_coef_map_id0 ?intr_eq0 ?ltr0_neq0 //.
  by move: B; case: (lead_coef p) => n.
rewrite -BpiE -(coefK Bpi) polyOver_poly // => l cl.
by rewrite coef_map Cint_int.
Qed.

(* Virer les 0 de chaque poly *)
Definition pre_alpha (k : 'I_s.+1) := [seq a <- (pre_alpha' k) | a != 0].

Lemma pre_alpha_int (j : 'I_s.+1) :
    {c : nat | ((\prod_(a <- (pre_alpha j)) ('X - a%:P))
      *+ c \is a polyOver Cint) && (c != 0%N)}.
Proof.
move: (alpha'_int j) => [c /andP[Heq c_neq0]].
exists c; rewrite c_neq0 andbT; move: Heq.
rewrite (bigID (fun i => i == 0)) /=.
rewrite (eq_bigr (fun i => 'X)); last first.
  by move=> i /eqP ->; rewrite polyC0 subr0.
rewrite big_const_seq; set nu := count _ _.
rewrite -Monoid.iteropE /=.
have -> : iterop nu *%R 'X 1 = 'X^nu  by [].
rewrite -mulrnAr big_filter => /polyOverP Hp; apply/polyOverP => i.
move: (Hp (nu + i)%N).
by rewrite coefXnM -{2}[nu]addn0 ltn_add2l  -{2}[nu]addn0 subnDl subn0.
Qed.

(* Multiplier les differents poly *)

Lemma eq_alpha_neq0 :
    {c : nat | ((\prod_(k < s.+1) \prod_(i <- (pre_alpha k)) ('X - i%:P))
      *+ c \is a polyOver Cint) && (c != 0)%N}.
Proof.
move: pre_alpha_int => H.
pose c k := proj1_sig (H k).
exists (\big[muln/1%N]_(k < s.+1) c k); apply/andP; split.
  rewrite -mulr_natr.
  have Hre r f : (\prod_(i <- r) (f i))%:R = \prod_(i <- r) ((f i)%:R).
    elim: r => [|a r ihr]; first by rewrite !big_nil.
    by rewrite big_cons big_cons -ihr natrM.
  rewrite Hre /= -big_split;apply: rpred_prod => k _.
  rewrite /= mulr_natr /c.
  by case: (H k)=> x /= /andP[].
rewrite -lt0n; apply: prodn_gt0 => k; rewrite /c lt0n.
by case: (H k)=> x /= /andP[].
Qed.

Lemma prod_beta_sum_alpha : 
  \prod_(b <- betaseq) (1 + (Cexp b)) = 
     \sum_(j < s.+1) \sum_(a <- pre_alpha' j) (Cexp a) + 1.
Proof.
rewrite polysym_Cexpa; congr (_  + _ ).
apply: eq_bigr => k _; rewrite big_map  big_filter.
by apply: eq_bigl => h; rewrite /= inE.
Qed.

(* Obtention des equations *)
Lemma eq_alpha_0 :
    \sum_(j < s.+1) \sum_(a <- pre_alpha' j) (Cexp a) + 1 = 0.
Proof.
by rewrite -prod_beta_sum_alpha prod_betaseq_Cexp.
Qed.

Lemma eq_alpha :
    (\sum_(j < s.+1) (size [seq a <- (pre_alpha' j) | a == 0]%R)).+1%:R +
      (\sum_(j < s.+1) \sum_(a <- pre_alpha j) (Cexp a)) = 0.
Proof.
rewrite -[X in _ = X]eq_alpha_0 -addn1 natrD -addrA [X in _ + X = _]addrC.
rewrite addrA /=; congr (_ + _).
have Hre r f : (\sum_(i <- r) (f i))%:R = \sum_(i <- r) ((f i)%:R).
  by rewrite natr_sum.
rewrite Hre -big_split /=; apply: eq_bigr => k _.
rewrite [X in _ = X](bigID (fun i => i == 0)) /=.
rewrite [X in _ = X + _](eq_bigr (fun i => 1)); last first.
  by move=> i /eqP ->; rewrite Cexp0.
rewrite /pre_alpha big_filter; congr(_ + _).
rewrite -sum1_size Hre big_filter.
by apply: eq_bigr => i _.
Qed.

(* n > 0 *)
Lemma ipi_pre_alpha :
    'i * PI%:C \in (pre_alpha ord0).
Proof.
suff H : 'i * PI%:C \in (pre_alpha' ord0).
  rewrite /pre_alpha mem_filter H andbT; apply: mulf_neq0.
    by rewrite eq_complex /= eq_refl andTb oner_neq0.
  rewrite eq_complex /= eq_refl andbT /GRing.zero /=.
  by move: PI_neq0; apply: (introN eqP).
move: iPi_in_betaseq => /nthP Hipi.
move: (Hipi 0) => {Hipi} [i Hi Heq].
have Hordi : (i < s.+1)%N by move: Hi; rewrite size_beta.
have -> : 'i * PI%:C = \sum_(i0 in [set (Ordinal Hordi)]) betatuple`_i0.
  by rewrite big_set1 Heq. 
by apply: image_f; rewrite inE /=; rewrite cards1.
Qed.

Definition all_alpha := flatten [seq pre_alpha k | k <- enum 'I_s.+1].

(* **** Definitions pour mpolytrans **** *)
(* n is r in Niven's paper *)
Definition n := size all_alpha.

Fact n_neq0 : n != 0%N.
Proof.
rewrite -lt0n /n -has_predT; apply/hasP.
exists ('i * PI%:C)=>//.
apply/flatten_mapP; exists ord0.
  by rewrite mem_enum.
by apply: ipi_pre_alpha.
Qed.

(* k is the same as in Niven, formula 9 *)
Definition k :=
    ((\sum_(j < s.+1) (size [seq a <- (pre_alpha' j) | a == 0]%R)).+1 : int).

Fact k_neq0 : k != 0.
Proof. by rewrite /k -lt0n. Qed.

Definition gamma : nat -> int := (fun n => 1).

Fact gamman_neq0 : gamma (n.-1) !=0.
Proof. by []. Qed.

Definition alpha : nat -> complexR := nth 0 all_alpha.

Fact  alpha_neq0 (i : 'I_n): alpha i != 0.
Proof.
have : all_alpha`_i \in all_alpha.
  by apply: mem_nth; move: (ltn_ord i).
by move/flatten_mapP => [j Hj ]; rewrite /pre_alpha mem_filter=>/andP [].
Qed.

Definition cT := sval eq_alpha_neq0.

Fact c_neq0 : cT != 0%N.
Proof. 
by move: (svalP eq_alpha_neq0) => /andP[ _]; rewrite /cT.
Qed.

Lemma all_alpha_Cint :
    \prod_(i <- (in_tuple all_alpha)) ('X - i%:P) *+ cT \is a polyOver Cint.
Proof.
move: (svalP eq_alpha_neq0) => /andP[H _].
by rewrite big_flatten /= big_map /= -filter_index_enum big_map /= /cT.
Qed.

Lemma eq_caract :
    k%:~R + (\sum_(0 <= i < n) (gamma i)%:~R * Cexp (alpha i)) = 0.
Proof.
    rewrite /k /gamma /n.
set a := _.+1.
have -> : a%:~R = (a%:R : complexR) by rewrite natzc.
rewrite (eq_bigr (fun i : nat => Cexp (alpha i))); last first.
  by move=> i _; rewrite mul1r.
rewrite /alpha -(big_nth 0 (fun i => true) (fun i => Cexp i)).
by rewrite big_flatten big_map -filter_index_enum big_map eq_alpha.
Qed.

Fact T_Cint : (@T n cT alpha) \is a polyOver Cint.
Proof.
rewrite /T /cT /n /alpha.
case: eq_alpha_neq0=> x /= /andP[H _].
rewrite -(big_mkord (fun i => true) (fun i => 'X - (nth 0 all_alpha i)%:P)).
rewrite -(big_nth 0 (fun i => true) (fun i => 'X - i%:P)).
by rewrite big_flatten big_map  -filter_index_enum big_map.
Qed.

Notation T := (T n cT alpha).
Notation Fpd := (Sd n k gamma cT alpha 0).
Hint Resolve n_neq0 c_neq0 T_Cint.
Notation p_gt0 := (p_gt0 n k gamma cT alpha).
Notation prime_p := (prime_p n k gamma cT alpha).
Notation leq_pk := (leq_pk n k gamma cT alpha).
Definition G := (G n k gamma cT alpha).
Definition  p := (p n k gamma cT alpha).
(**** * Combining the two parts * ****)

Lemma pi_transcendental_aux : False.
Proof.
apply: (main_contradiction n_neq0 k_neq0 c_neq0 alpha_neq0 T_Cint eq_caract).
rewrite {1}/gamma.
set pT := p.
rewrite (eq_bigr (fun i => G.[alpha i] * (cT ^ (n * pT))%:R)); first last.
  by move=> i _; rewrite  mul1r mulr_natr.
have G_polyOver : G \is a polyOver Cint.
  apply: rpred_sum => i _.
  by apply: polyOver_derivn; apply: pre_G_Cint.
have G_size : (size G = n * pT)%N.
  by rewrite (size_G n_neq0 _ _ c_neq0) /pT mulnC.
pose nu := n.-1.
have Hn : n = nu.+1 by rewrite prednK // lt0n n_neq0.
have Hord (i : 'I_n) : nat_of_ord (@inord nu i) = nat_of_ord i.
  by rewrite inordK // /nu prednK // lt0n n_neq0.
pose Qmj (j : 'I_nu.+1) := \sum_(i < (n * pT)) G`_i *: 'X_j ^+ i.
pose Qm := \sum_(i < nu.+1) (Qmj i).
rewrite -big_distrl /= mulr_natr.
have Htuple : size (all_alpha) == nu.+1.
  rewrite /nu /n prednK //.
  by move: n_neq0; rewrite /n -lt0n.
have -> : \sum_(0 <= i < n) G.[alpha i] = Qm.@[Tuple Htuple].
  rewrite /Qm raddf_sum Hn big_mkord /=; apply: eq_bigr => i _.
  rewrite horner_coef /Qmj raddf_sum /= G_size; apply: eq_bigr => j _.
  rewrite -mul_mpolyC mevalM mevalC rmorphX /= mevalX.
  by rewrite (tnth_nth 0) /= /alpha.
case: (boolP (Qm == 0)) => [/eqP -> | Qm_neq0].
  by rewrite meval0 mul0rn.
have Qm_mpolyOver : Qm \is a mpolyOver Cint.
  rewrite /Qm; apply: rpred_sum => i _.
  rewrite /Qmj; apply: rpred_sum => j _.
  rewrite -mul_mpolyC; apply: rpredM.
    by rewrite mpolyOverC; move: G_polyOver => /polyOverP QP; apply: QP.
  by apply: rpredX; apply: mpolyOverX.
have Qm_sym : Qm \is symmetric.
  apply/issymP => s; rewrite /Qm raddf_sum /=.
  rewrite [X in _ = X](reindex s) /=; last first.
    exists (s^-1)%g => x Hx.
      by rewrite permK.
    by rewrite permKV.
  apply: eq_bigr => i _; rewrite /Qmj raddf_sum /=; apply: eq_bigr => j _.
  rewrite msymZ rmorphX /= ; congr (fun x => G`_j *: x ^+j).
  by rewrite /msym /= mmapX mmap1U.
have Qm_size : (msize Qm <= (n * pT))%N.
  rewrite /Qm.
  rewrite (@leq_trans (\max_(i < nu.+1) msize (Qmj i))) //.
    by apply:msize_sum.
  apply/bigmax_leqP => i _.
  rewrite /Qmj.
  rewrite (@leq_trans (\max_(i0 < n * pT) msize (G`_i0 *: 'X_i ^+ i0))) //.
    by apply:msize_sum.
  apply/bigmax_leqP => j _.
  case:(boolP (G`_j  == 0)).
    by move/eqP ->; rewrite scale0r msize0.
   move/msizeZ => ->.
  rewrite mpolyXn mmeasureE msuppX big_seq1 /=.
  have -> : (mdeg (U_(i) *+ j) = (mdeg U_(i)) * j)%N.
    elim: (nat_of_ord j) => [|nj ihnj].
      by rewrite mulm0n mdeg0 muln0.
    by rewrite mulmS mdegD ihnj -mulnS.
  by rewrite mdeg1 mul1n.
move: (mpolysym Qm_mpolyOver Qm_sym Qm_neq0)
  => [R /and3P[/eqP Req Rsize  /mpolyOverP R_mpolyOver]].
rewrite Req.
set lq := [tuple _ | _ < _].
set cs := Tuple Htuple.
pose Htu := [seq i.@[cs] | i <- lq].
have Htusize : size Htu == nu.+1.
  by rewrite /Htu size_map size_tuple eq_refl.
have <- : R.@[Tuple Htusize]=(R \mPo lq).@[cs].
  rewrite {1}/meval /mmap /comp_mpoly /mmap raddf_sum.
  apply: eq_bigr => m _ /=.
  rewrite mevalM mevalC; congr (_ * _).
  rewrite /mmap1 rmorph_prod /=; apply: eq_bigr => i _ /=.
  rewrite rmorphX /=; congr (fun x => x ^+ m i).
  rewrite !(tnth_nth 0)  (nth_map 0) // size_tuple.
  by apply: ltn_ord.
have HR_size : (msize R <= n * pT)%N.
  apply: (leq_trans (leq_msize_meight _)).
  by apply: (leq_trans Rsize Qm_size).
rewrite /meval /mmap -mulr_natr big_distrl /= big_seq.
apply: rpred_sum => i /msize_mdeg_lt Hi; rewrite -mulrA; apply: rpredM.
  by apply: R_mpolyOver.
rewrite /mmap1 Hn.
have Hdegi : (mdeg i < (n * pT))%N.
  by apply: (leq_trans Hi HR_size).
have -> : (cT ^ (nu.+1 * pT))%:R = 
           (\prod_(i0 < nu.+1) (cT ^ (i i0))%:R) * (cT ^ (n * pT - mdeg i))%:R.
  move=> t.
  have HreM r f : (\prod_(i <- r) (f i))%:R = \prod_(i <- r) ((f i)%:R).
    elim: r => [|a r ihr]; first by rewrite !big_nil.
    by rewrite big_cons big_cons -ihr natrM.
  rewrite -HreM -expn_sum -natrM -expnD -mdegE subnKC ?Hn //=.
    by apply: ltnW; move: Hdegi; rewrite Hn.
rewrite mulrA; apply: rpredM; last by apply: Cint_Cnat; apply: Cnat_nat.
rewrite -big_split /=; apply: rpred_prod => j _.
rewrite natrX -exprMn; apply: rpredX; rewrite (tnth_nth 0) mulr_natr.
have -> : (Tuple Htusize)`_j = 's_(nu.+1, j.+1).@[Tuple Htuple].
  rewrite (nth_map 0) /cs /lq /=; last first.
    by rewrite size_map size_enum_ord.
  by rewrite (nth_map ord0) ?nth_ord_enum // size_enum_ord.
move: all_alpha_Cint => /=.
have -> : \prod_(i0 <- all_alpha) ('X - i0%:P) =
              \prod_(i0 <- Tuple Htuple) ('X - i0%:P) by [].
move => /polyOverP Halpha.
have Hordk : (j.+1 < nu.+2)%N  by rewrite ltnS.
move: (Halpha (nu.+1 - (Ordinal Hordk))%N).
by rewrite coefMn mroots_coeff /= -mulrnAr rpredMsign.
Qed.

End Preliminary.

Section Trans_pi.

Theorem pi_transcendental : ~algebraicOver ratr PI%:C.
Proof. by apply: pi_transcendental_aux. Qed.

(* To verify the development, the command

  Print Assumptions pi_transcendental.

   will list all the assumptions that have been used in
   the formal proof.  These assumptions are only the
   definitions of R, and the axioms for classical logic. *)

End Trans_pi.
